## pluginhandler

from baseclasses import handler
from string import join
import sys,os
import imp			## module loading
import traceback		## module import + function execution
import threading
import time
import re

class pluginhandler(handler):
	__default_functions = {} ## default unctions as name/function-reference key/value-pairs
	__functions = {}
	__commands = {}
	__plugins = {}
	__threads = {} ## user based command execution threads
	__access_time = {} ## user based access time for dos detection

	def __init__(self):
		self.__exit = False

		## register default functions
		self.__default_functions['log'] = self._default_log
		self.__default_functions['get_userlevel'] = self._default_get_userlevel
		self.__default_functions['read_config'] = self._default_read_config
		self.__default_functions['get_config'] = self._default_get_config
		self.__default_functions['get_plugin_config'] = self._default_get_plugin_config

		## config
		self._plugindir = '.'
		self.__default_config = {	'startup': '[]',
						'min_access_time': '1.0',
						'ignore_time': '10.0',
						'thread_error': 'thread error'}
		self._configure()

		## threading
		self.__plugin_rlock = threading.RLock()

	def __getattr__(self, name):
		if self.__functions.has_key(name):
			return self.__functions[name]['function']
		if self.__default_functions.has_key(name):
			return self.__default_functions[name]
		raise AttributeError, name


	####################################################################################################
	## set config parameters for ph
	def _configure(self):
		self.__config = self.get_config('pluginhandler', self.__default_config)
		self.__min_access_time = float(self.__config['min_access_time'])
		self.__ignore_time = float(self.__config['ignore_time'])
		
	####################################################################################################
	## plugin handling

	def _load(self, name):
		"""load plugin"""
		if self.__plugins.has_key(name):
			return self.log("error loading: plugin %s already exists" % name, 'error')
			
		if not os.path.isfile(self._plugindir+'/'+name+'.py'):
			return self.log("error loading: plugin %s doesn't exist" % name, 'error')

		startloadtime = time.clock()

		self.__plugin_rlock.acquire()
		fp, pathname, description = imp.find_module(name, [self._plugindir])
		try:
			module = imp.load_module(name, fp, pathname, description)
			self.__plugins[name]={'name': name, 'module': module,
				'obj': eval('module.'+name+'(ph=self)')}
		except ImportError, e:
			self.__plugin_rlock.release()
			error = self.log("error importing module %s: %s" % (name, str(e)), 'error')
			if fp:
				fp.close()
			return error
		except Exception, e:
			self.__plugin_rlock.release()
			error = self.log("error loading module %s: %s" % (name, str(e)), 'error')
			self.log("--- stack traceback ---\n%s-----------------------" % join(traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback), ''), 'traceback')
			self._unload(name)
			if fp:
				fp.close()
			return error
		if fp:
			fp.close()

		self.__plugin_rlock.release()

		## init
		try:
			self.__plugins[name]['obj']._on_load()
		except Exception, e:
			error = self.log("error calling _on_load() in plugin %s: %s" % (name, str(e)), 'error')
			self.log("--- stack traceback ---\n%s-----------------------" % join(traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback), ''), 'traceback')
			self._unload(name)
			return error
			

		endloadtime = time.clock()
		self.log("* successfully loaded %s in %.3fs" % (name, endloadtime-startloadtime), 'module')

		return None

	def _unload(self, name):
		"""unload plugin"""
		if not self.__plugins.has_key(name):
			return self.log("error unloading: plugin %s is not loaded" % name, 'error')

		module = self.__plugins[name]['obj']
		
		## wait for input threads
		if module._type == 'input':
			module._exit = True
			self.__plugins[name]['thread'].join()
			
		self.__plugin_rlock.acquire()
		
		## uninit
		try:
			module._on_unload()
		except Exception, e:
			self.log("error calling _on_unload() in plugin %s: %s" % (name, str(e)), 'error')
			self.log("--- stack traceback ---\n%s-----------------------" % join(traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback), ''), 'traceback')


		## remove commands
		for cmd in self.__commands.keys():
			if self.__commands[cmd]['module'] is module:
				self.__commands.pop(cmd)

		## remove commands
		for cmd in self.__functions.keys():
			if self.__functions[cmd]['module'] is module:
				self.__functions.pop(cmd)

		## remove plugin		
		self.__plugins.pop(name)
		self.__plugin_rlock.release()
		self.log("* successfully unloaded %s" % name, 'module')

		return None
	
	def _register_command(self, plugin, name, function, help='', args={}):
		"""register command"""
		if self.__commands.has_key(name):
			self.log("error registering cmd: %s already exists in plugin %s" % (name, self.__commands[name]['module'].name), 'error')
		else:
			self.log("registering command /%s from plugin %s" % (name, plugin._name), 'module')
			self.__commands[name] = {'name': name, 'module': plugin, 'function': function,
				'help': help, 'args': args}
		

	def _register_function(self, plugin, name, function):
		"""register a function as member of ph"""
		if self.__functions.has_key(name):
			return self.log("function %s already exists from module %s" % (name, plugin._name), 'error')
		self.__functions[name] = { 'name':name, 'module':plugin, 'function':function }
	
	def _unload_all(self):
		for module in self.__plugins.keys():
			self._unload(module)

	####################################################################################################
	## default functions -- overridden by core plugins

	def _default_log(self, message, type='info'):
		print message
		return message
	
	def _default_get_userlevel(self, uid):
		"""return userlevel which can be seen as an access permission level
			0 is highest"""
		return 0
	
	def _default_read_config(self, filename):
		pass
	
	def _default_get_config(self, section, default = {}):
		return default
	
	def _default_get_plugin_config(self, plugin, default = {}):
		return self.get_config(plugin._name, default)
	


	####################################################################################################
	## info functions

	def _get_commandlevel(self, name):
		cmd = self.__commands[name]
		if cmd.has_key('args') and cmd['args'].has_key('level'):
			return cmd['args']['level']
		return 100
	
	def _get_plugin_info(self):
		names = {}
		for p in self.__plugins.keys():
			plugin = self.__plugins[p]
			names[plugin['name']] = {'type': plugin['obj']._type, 'desc': plugin['obj']._desc }
		return names
	
	def _get_command_info(self, wanted_cmds=[]):
		if not wanted_cmds:
			all_cmds = self.__commands.keys()
		else:
			all_cmds = wanted_cmds
		cmds = {}
		for c in all_cmds:
			if self.__commands.has_key(c):
				cmd = self.__commands[c]
				longhelp = None
				if cmd['args'].has_key('longhelp'):
					longhelp = cmd['args']['longhelp']
				cmds[c] = {	'module': cmd['module']._name,
						'level': self._get_commandlevel(c),
						'help': cmd['help'],
						'longhelp': longhelp }
		return cmds

	####################################################################################################
	## command processing

	def __start_thread(self, func, args):
		user = args['msg'].get_uid()
		if self.__threads.has_key(user) and str(type(self.__threads[user])) == "<class 'threading.Thread'>" and self.__threads[user].isAlive():
			args['msg'].send_back(self.__config['thread_error'])
			return
		self.__threads[user] = threading.Thread(target=func, args=(args,))
		self.__threads[user].setName(args['cmd'][0])
		self.__threads[user].start()
		
	def _process_request(self, msg, threaded=False, dos_detection=False):
		text = msg.get()
		if len(text) <= 0:
			return
		if text[0] != '/':
			text = '/msg ' + text

		cmd = re.split('[\s\t\r\n]+', text[1:])
		command = cmd[0]

		## dos detection
		if dos_detection:
			now = time.time()
			user = msg.get_uid()
			if self.__access_time.has_key(user):
				if self.__access_time[user].has_key('ignore') and (self.__access_time[user]['ignore'] + self.__access_time[user]['last_access']) > now:
					## ingoring request
					return
				if abs(now - self.__access_time[user]['last_access']) < self.__min_access_time:
					## user is too fast. possible dos attack
					self.__access_time[user] = {'last_access': now, 'ignore': self.__ignore_time}
					msg.send_back("dos detected. you are being ignored for %.1f seconds." % self.__ignore_time)
					return
			self.__access_time[user] = {'last_access': now}

		## logging
		self.log('<' + msg.get_uid() + '> ' + join(re.split('\r?\n', text), '\n\t'), 'inmessage')
		
		## check if we actually have such a function
		if not self.__commands.has_key(cmd[0]):
			msg.send_back('error. try /help')
			return

		## check if user is allowed to execute command
		if self._get_commandlevel(command) < self.get_userlevel(msg.get_uid()):
			msg.send_back('error. you have insufficient permissions to execute this command.')
			return

		## execute command
		try:
			func = self.__commands[command]['function']
			args = {'msg': msg, 'cmd': cmd}
			if threaded:
				self.__start_thread(func, args)
			else:
				func(args)
		except Exception, e:
			self.log("error executing command %s: %s" % (cmd[0], str(e)), 'error')
			self.log("--- stack traceback ---\n%s-----------------------" % join(traceback.format_exception(sys.exc_type, sys.exc_value, sys.exc_traceback), ''), 'traceback')
			msg.send_back('error executing function')
		
	####################################################################################################
	## quit functions

	def _quit(self):
		## set exit flags
		self.__exit = True
		for name in self.__plugins.keys():
			self.__plugins[name]['obj']._exit = True

	def __quit(self):
		## wait for command threads
		for cm in self.__threads:
			if self.__threads[cm].isAlive() and self.__threads[cm] is not threading.currentThread():
				self.__threads[cm].join()

		## unload non-input modules
		for module in self.__plugins.keys():
			if not self.__plugins[module]['obj']._type == 'input':
				self._unload(module)

		## unload input modules
		self._unload_all()

		sys.exit(0)
	
	####################################################################################################
	## startup

	def _load_startup_plugins(self):
		plugins = eval(self.__config['startup'])
		for plugin in plugins:
			self._load(plugin)
			

	####################################################################################################
	## input thread handling

	def __input_process_thread(self, plugin):
		"""input thread main function"""
		self.log("input processing started for " +plugin._name)
		while not self.__exit and not plugin._exit:
			plugin.process()
			
	def _start_input_process(self, plugin_name):
		"""start thread for input plugin"""
		plugin = self.__plugins[plugin_name]['obj']
		fkt = self.__input_process_thread
		thread = threading.Thread(target=fkt, args=(plugin,))
		thread.setName(plugin_name)
		self.__plugins[plugin_name]['thread'] = thread
		thread.start()
		

	####################################################################################################
	## run - the main function
	
	def _run(self):
		while not self.__exit:
			self.__plugin_rlock.acquire()
			if not self.__exit:
				for name in self.__plugins.keys():
					plugin = self.__plugins[name]
					if plugin['obj']._type == 'input':
						if not plugin.has_key('thread') or not plugin['thread'].isAlive():
							if not plugin['obj']._exit:
								self._start_input_process(name)
			self.__plugin_rlock.release()
			if not self.__exit:
				time.sleep(10)

		self.__quit()
		
		pass


	####################################################################################################
	## send messages
	
	def __find_handler(self, inputtype):
		for name in self.__plugins.keys():
			if self.__plugins[name]['obj']._type == 'input':
				if self.__plugins[name]['obj']._inputtype == inputtype:
					return self.__plugins[name]['obj']
		return None
		
	def send_message(self, id, message):
		"""send a message to a user id of the form 'handler:id' """
		pos = id.find(':')
		if pos < 0:
			return "error: invalid user id"
		intype = id[:pos]
		name = id[pos+1:]
		handler = self.__find_handler(intype)

		if not handler:
			return "error: no handler found for type " + intype
		
		return handler._send(name, message)

	####################################################################################################
	## user functions

	def ignore_user(self, user, seconds):
		self.__access_time[user] = {'last_access': time.time(), 'ignore': seconds}

	def user_is_available(self, user):
		pos = user.find(':')
		if pos < 0:
			return False
		intype = user[:pos]
		name = user[pos+1:]
		handler = self.__find_handler(intype)
	
		if not handler:
			return False
		return handler._is_available(name)

	def user_last_access(self, user):
		if self.__access_time.has_key(user) and self.__access_time[user].has_key('last_access'):
			return self.__access_time[user]['last_access']
		return None
			
