# -*- coding: iso8859-15 -*-
##
## jabberif AI plugin
##	this plugin provides a kind of artificial intelligence using
##	pattern matching rules from AIML files
##
##	the <system/> tag from the aiml tag library has been overloaded
##	in order to provide a jabberif command interface
##
##
##	written by Ben Fuhrmannek <bef@erlangen.ccc.de>
##
##

from baseclasses import output_plugin,input_plugin
from helpers import message
import sys,os
import aiml
from string import join
import marshal
import copy
import time
import re
import threading



############################################################################################################

class ai_command_cache(input_plugin):
	"""dummy input plugin to handle command redirection from within the ai"""
	def __init__(self, ph):
		self.ph = ph

	def process(self):
		pass

	def redirect_command(self, id, cmd):
		pos = id.find(':')
		intype = id[:pos]
		name = id[pos+1:]

		self.__message = ''

		msg = message(intype, self, name, cmd)
		self.ph._process_request(msg=msg, threaded=False, dos_detection=False)
	
	def _send_back(self, msg, message):
		self.__message = message

	def get_message(self):
		return self.__message

############################################################################################################

import xml.parsers.expat

class botpredparser:
	__bot_predicates = {}
	__tag_cache = []
	def __init__(self):
		pass
	def start_element(self, name, attrs):
		if len(self.__tag_cache) > 1:
			raise SyntaxError
		self.__tag_cache.append([name, attrs])
		if name == "bot" and not attrs.has_key("name"):
			raise SyntaxError
	def end_element(self, name):
		elem = self.__tag_cache.pop()
		if elem[0] != name:
			raise SyntaxError
	def char_data(self, data):
		if self.__tag_cache[len(self.__tag_cache)-1][0] == "bot":
			self.__bot_predicates[self.__tag_cache[len(self.__tag_cache)-1][1]['name']] = data

	def parse(self, filename):
		if not os.path.isfile(filename):
			raise IOError
	
		self.__bot_predicates = {}
		p = xml.parsers.expat.ParserCreate()

		p.StartElementHandler = self.start_element
		p.EndElementHandler = self.end_element
		p.CharacterDataHandler = self.char_data

		predfile = file(filename)
		p.Parse(predfile.read(), 1)
		predfile.close()
		return self.__bot_predicates

############################################################################################################

class ai_plugin(output_plugin):
	def __init__(self, ph):
		self.ph = ph
		self._name = 'ai_plugin'
		self._desc = 'imitate human-like responses'
		
		self.__brainevent = threading.Event()
		self.__brainevent.set()
		self.__sessionevent = threading.Event()
		self.__sessionevent.set()
		self.__brainlock = threading.RLock()
		self.__sessionlock = threading.RLock()

		self.__default_config = {	'human_delay':'0.1',
						'max_answer_time':'3.0',
						'ai_subst':'',
						'ai_botpreds':'',
						'aiml_commands':'',
						'aiml_source':'',
						'ai_brain':'/tmp/jabberif.brain',
						'ai_session':'/tmp/jabberif.session' }
		self._configure()

		## init AI
		self.__init_ai()

		## defs
		self.__umlauts = {ord(''):u'ss', ord(''):u'ae', ord(''):u'oe', ord(''):u'ue',
			ord(''):u'Ae', ord(''):u'Oe', ord(''):u'Ue'}
		#self.__umlauts = {'\xc3\x9f':'ss', '\xc3\xa4':'ae', '\xc3\xb6':'oe', '\xc3\xbc':'ue', '\xc3\x84':'Ae', '\xc3\x96':'Oe', '\xc3\x9c':'Ue'}
		
	
	def _configure(self):
		self.__config = self.ph.get_plugin_config(self, self.__default_config)
		self.__human_delay = float(self.__config['human_delay'])
		self.__max_answer_time = float(self.__config['max_answer_time'])
		self.__sessionfile = self.__config['ai_session']
		self.__brainfile = self.__config['ai_brain']
		self.__subsfile = self.__config['ai_subst']
		self.__botpredfile = self.__config['ai_botpreds']
	
	def _on_load(self):
		self.ph._register_command(self, 'msg', self.respond, 'ai robot')
		self.ph._register_command(self, 'aireloadbrain', self.reloadbrain, 'reload ai brain from brain file', {'level':5})
		self.ph._register_command(self, 'aireloadaiml', self.reloadaiml, 'reload ai brain from aiml file', {'level':5})
		self.ph._register_command(self, 'aisavebrain', self.savebrain, 'save ai brain into brain file', {'level':5})
		self.ph._register_command(self, 'aisavesession', self.savesession, 'save ai session', {'level':5})
		self.ph._register_command(self, 'ailoadsession', self.loadsession, 'load ai session', {'level':5})
		self.ph._register_command(self, 'aiviewsession', self.view_session, '[id] view session', {'level':30})
		self.ph._register_command(self, 'aiclearsession', self.clear_session, '[id] clear session', {'level':5})
		
	def _on_unload(self):
		self.__savesession(self.__sessionfile)
		
	####################################################################################################
	## init
	
	def __init_ai(self):
		aiml_source = self.__config['aiml_source']
		aiml_commands = self.__config['aiml_commands']
		brain = self.__brainfile
		old_wd = os.getcwd()
		os.chdir(os.path.dirname(aiml_source))
		
		## create kernel
		self.__k = aiml.Kernel()
		self.__system_override()

		## load substitution file
		if os.path.isfile(self.__subsfile):
			self.__k.loadSubs(self.__subsfile)

		## load brain
		if os.path.isfile(brain):
			self.__load_brain(brain)
		else:
			self.__load_aiml(aiml_source, aiml_commands)
			self.__k.saveBrain(brain)
		
		os.chdir(old_wd)

		sessionfilename = self.__sessionfile
		if os.path.isfile(sessionfilename):
			self.__loadsession(sessionfilename)
		
		self.__k.verbose(isVerbose=False)


	####################################################################################################
	## <system> aiml tag override
	
	def __system_override(self):
		## override <system> tag to handle jabberif commands
		self.__k._elementProcessors['system'] = self.__aiml_system_tag_handler

	def __reset_brain(self):
		self.__k.resetBrain()
		self.__system_override()

	def __aiml_system_tag_handler(self, elem, id):
		## get command
		cmd = ""
		for e in elem[2:]:
			cmd += self.__k._processElement(e, id)
		cmd = unicode(cmd).encode('latin-1', 'replace')
		
		## redirect command to our dummy input handler
		cache = ai_command_cache(self.ph)
		cache.redirect_command(id, cmd)
		message = cache.get_message()

		## ... maybe we do some postprocessing here
		## ... in the future.

		return message

	####################################################################################################
	## /msg command handler

	def respond(self, args):

		if not self.__sessionevent.isSet() or not self.__brainevent.isSet():
			args['msg'].send_back('database maintenance in progress. please try again later.')
			return
			
		input = join(args['cmd'][1:], ' ')
		user = args['msg'].get_uid()

		## substitute german umlauts
		#for umlaut in self.__umlauts:
		#	input.replace(umlaut, self.__umlauts[umlaut])
		input = unicode(input).translate(self.__umlauts)
		
		## generate response
		response = self.__k.respond(input, user)
		
		## human response time delay
		if self.__human_delay > 0.01:
			sleeptime = len(response) * self.__human_delay
			if sleeptime > self.__max_answer_time:
				sleeptime = self.__max_answer_time
			time.sleep(sleeptime)

		## send answer back
		args['msg'].send_back(response)
	

	####################################################################################################
	## load/save brain
	
	def __set_bot_predicates(self):
		bpp = botpredparser()
		__bot_predicates = bpp.parse(self.__botpredfile)
		for name,value in __bot_predicates.items():
			self.__k.setBotPredicate(name, value)
			
	def __load_brain(self, brain):
		self.__set_bot_predicates()
		self.__k.bootstrap(brainFile = brain)
	
	def __load_aiml(self, aiml_source, aiml_commands = ''):
		self.__set_bot_predicates()
		self.__k.bootstrap(learnFiles = aiml_source, commands = aiml_commands)
	
	def reloadbrain(self, args):
		brain = self.__brainfile
		
		self.__brainlock.acquire()
		self.__brainevent.clear()

		try:
			if os.path.isfile(brain):
				self.__reset_brain()
				self.__load_brain(brain)

				## reload session
				self.__loadsession(self.__sessionfile)

				args['msg'].send_back('successfully reloaded brain')
			else:
				args['msg'].send_back('error: cannot find brain file '+brain)
		finally:
			self.__brainevent.set()
			self.__brainlock.release()
		
	def reloadaiml(self, args):
		aiml_source = self.__config['aiml_source']
		aiml_commands = self.__config['aiml_commands']
		
		self.__brainlock.acquire()
		self.__brainevent.clear()
		
		try:
			if os.path.isfile(aiml_source):
				old_wd = os.getcwd()
				os.chdir(os.path.dirname(aiml_source))
				self.__reset_brain()
				self.__load_aiml(aiml_source, aiml_commands)
				os.chdir(old_wd)

				## reload session
				self.__loadsession(self.__sessionfile)
				
				args['msg'].send_back('successfully reloaded brain')
			else:
				args['msg'].send_back('error: cannot find aiml file '+aiml_source)
		finally:
			self.__brainevent.set()
			self.__brainlock.release()
	
	def savebrain(self, args):
		self.__brainlock.acquire()
		self.__brainevent.clear()
		try:
			brain = self.__brainfile
			self.__k.saveBrain(brain)
			args['msg'].send_back('brainfile saved')
		finally:
			self.__brainevent.set()
			self.__brainlock.release()


	####################################################################################################
	## session
		
	def __savesession(self, sessionfilename):
		self.__sessionlock.acquire()
		self.__sessionevent.clear()
		try:
			session = self.__k.getSessionData()
			sessionfile = file(sessionfilename, "wb")
			marshal.dump(session, sessionfile)
			sessionfile.close()
		finally:
			self.__sessionevent.set()
			self.__sessionlock.release()
	
	def savesession(self, args):
		self.__savesession(self.__sessionfile)
		args['msg'].send_back('session saved')

	def __loadsession(self, sessionfilename):
		self.__sessionlock.acquire()
		self.__sessionevent.clear()
		try:
			sessionfile = file(sessionfilename, "rb")
			session = marshal.load(sessionfile)
			sessionfile.close()
			#for sessionid,values in session.items():
			#	for pred,value in values:
			#		self.__k.setPredicate(pred, value, sessionid)
			self.__k._sessions = copy.deepcopy(session)
		finally:
			self.__sessionevent.set()
			self.__sessionlock.release()
		
	def loadsession(self, args):
		self.__loadsession(self.__sessionfile)
		args['msg'].send_back('session loaded')

	def view_session(self, args):
		if not self.__sessionevent.isSet():
			args['msg'].send_back('database maintenance in progress. please try again later.')
			return

		ids = args['cmd'][1:]
		out = []
		if len(ids) < 1: ## list all avail. ids
			out = self.__k.getSessionData().keys()
			out.sort()
		else: ## show wanted ids
			data = self.__k.getSessionData(ids[0])
			for key in data.keys():
				if not key[:1] == '_':
					values = re.split('[\n\r\s\t]+', data[key])
					value = join(values, ' ')
					if len(join(values, '')) > 0:
						out.append(key + ': '+ value)

		args['msg'].send_back('\n'+join(out, '\n'))

	def clear_session(self, args):
		if not self.__sessionevent.isSet():
			args['msg'].send_back('database maintenance in progress. please try again later.')
			return

		ids = args['cmd'][1:]
		if len(ids) < 1: ## list all avail. ids
			args['msg'].send_back('syntax error. see /help '+args['cmd'][0])
			return
		self.__k._deleteSession(ids[0])
		args['msg'].send_back('cleared.')

