##
## jabberif randomtalk plugin
##	randomly start chats with people
##
##	written by Ben Fuhrmannek <bef@erlangen.ccc.de>
##

from baseclasses import output_plugin
from string import join,strip
import time
import re
import random
import threading
import copy

class randomtalk_plugin(output_plugin):
	def __init__(self, ph):
		self.ph = ph
		self._name = 'randomtalk_plugin'
		self._desc = 'start chats'

		self.__default_config = {	'min_wait':'60',
						'max_wait':'7200',
						'messages':'["hi there."]',
						'activate_on_startup':'no',
						'users':'raw:stdin',
						'autorestart':'yes',
						'check_online_status':'yes',
						'check_last_accessed_time':'3600'}
		self._configure()
		
		## init
		self.__timer = None
		self.__endtime = None
		if self.__config['activate_on_startup'] in ['on', '1', 'yes', 'true', 'go']:
			self.__start()

	
	def _configure(self):
		self.__config = self.ph.get_plugin_config(self, self.__default_config)
		self.__min_wait = int(self.__config['min_wait'])
		self.__max_wait = int(self.__config['max_wait'])
		self.__users = re.split('[,\n\s\t]+', self.__config['users'])
		self.__check_users()
		self.__messages = eval(self.__config['messages'])
		self.__set_autorestart(self.__config['autorestart'])
		
		self.__check_online = False
		if self.__config['check_online_status'] in ['on', '1', 'yes', 'true', 'go']:
			self.__check_online = True
		self.__check_access = int(self.__config['check_last_accessed_time'])
	
	def _on_load(self):
		self.ph._register_command(self, 'rt', self.randomtalk, '[start|stop|help] random talk', {'level':20})
	
	def _on_unload(self):
		self.__stop()
	
	
	####################################################################################################

	def __check_users(self):
		for id in range(len(self.__users)):
			if not self.__check_user(self.__users[id]):
				self.__users.pop(id)
	def __check_user(self, name):
		parts = re.split(':', name)
		if len(parts) < 2 or len(parts[0]) < 1 or len(parts[1]) < 1:
			return False
		return True
		
	####################################################################################################
	
	def __find_user(self):
		userlist = copy.copy(self.__users)
		random.shuffle(userlist)
		if self.__check_access == 0 and not self.__check_online:
			return userlist[0]

		selected = None
		for user in userlist:
			if self.__check_online and self.ph.user_is_available(user):
				selected = user

			if self.__check_access > 0:
				last_access = self.ph.user_last_access(user)
				if last_access == None or self.__check_access < time.time() - last_access:
					selected = user
				else:
					selected = None

			if selected:
				return selected

		return selected
		
	def __action(self):
		user = self.__find_user()
		if user:
			message = self.__messages[random.randint(0, len(self.__messages)-1)]
			self.ph.send_message(user, message)
			self.ph.log('{'+user+'} ' + message, 'outmessage')
			self.__stop()
		if self.__autorestart:
			self.__start()

	def __set_autorestart(self, value):
		if value in ['on', '1', 'yes', 'true', 'go']:
			self.__autorestart = True
			return "on"
		else:
			self.__autorestart = False
			return "off"
		

	def __help(self):
		return "possible params: start, stop, help, status, list, add, set [minwait|maxwait|autorestart]"
	
	def __status(self):
		if not self.__timer:
			return "no timer started"
		return "timer ends in %ds" % int(self.__endtime - time.time())

	def __start(self):
		## stop timer first
		if self.__timer:
			self.__timer.cancel()

		## find random params
		waittime = random.randint(self.__min_wait, self.__max_wait)
		self.__timer = threading.Timer(waittime, self.__action)
		self.__endtime = time.time() + waittime
		self.__timer.start()
		return "timer started: %ds" % int(self.__endtime - time.time())

	def __stop(self):
		if self.__timer:
			self.__timer.cancel()
			self.__timer = None
			self.__endtime = None
			return "timer stopped"
		else:
			return "no timer available"
	
	def __list_users(self):
		out = ['\n']
		for id in range(len(self.__users)):
			out.append("%02d %s" % (id, self.__users[id]))
		return join(out, '\n')
		
	def __add_users(self, users):
		out = []
		for user in users:
			if self.__check_user(user):
				self.__users.append(user)
				out.append(user)
		return "added the following users: "+join(out, ', ')
		
	def __del_users(self, users):
		out = []
		for user in users:
			for myid in range(len(self.__users)):
				if user == self.__users[myid]:
					self.__users.pop(myid)
					out.append(user)
		return "removed the following users: "+join(out, ', ')

	def __set_max_wait(self, t):
		try:
			value = int(t)
		except:
			return "wrong time"
		self.__max_wait = value
		return "new max wait time is "+t+"s"

	def __set_min_wait(self, t):
		try:
			value = int(t)
		except:
			return "wrong time"
		self.__min_wait = value
		return "new min wait time is "+t+"s"

	def randomtalk(self, args):
		params = args['cmd'][1:]

		out = ''
		if not params or params[0] == 'status':
			args['msg'].send_back(self.__status())
		elif params[0] == 'help':
			args['msg'].send_back(self.__help())
		elif params[0] == 'start':
			args['msg'].send_back(self.__start())
		elif params[0] == 'stop':
			args['msg'].send_back(self.__stop())
		elif params[0] == 'list':
			args['msg'].send_back(self.__list_users())
		elif params[0] == 'del':
			args['msg'].send_back(self.__del_users(params[1:]))
		elif params[0] == 'add':
			args['msg'].send_back(self.__add_users(params[1:]))
		elif params[0] == 'set' and len(params) == 3 and params[1] == 'minwait':
			args['msg'].send_back(self.__set_min_wait(params[2]))
		elif params[0] == 'set' and len(params) == 3 and params[1] == 'maxwait':
			args['msg'].send_back(self.__set_max_wait(params[2]))
		elif params[0] == 'set' and len(params) == 3 and params[1] == 'autorestart':
			args['msg'].send_back(self.__set_autorestart(params[2]))
		elif params[0] == 'set' and len(params) == 1:
			args['msg'].send_back("minwait=%d, maxwait=%d"%(self.__min_wait, self.__max_wait))
		else:
			args['msg'].send_back('error. try /'+args['cmd'][0]+' help')
			

